@@ -42,8 +42,6 @@ from .parser import (
     parse_field,
     parse_literal,
     parse_query,
-    skip_optimizations,
-    strict_field_schema,
     extract_query_terms,
 )
 from .schema import Schema
@@ -68,7 +66,7 @@ from .walkers import (
     Walker,
 )
 
-__version__ = '0.8.7'
+__version__ = '0.9.3'
 __all__ = (
     "__version__",
     "AnalyticOutput",
@@ -124,6 +122,4 @@ __all__ = (
     "save_analytic",
     "save_analytics",
     "save_dump",
-    "skip_optimizations",
-    "strict_field_schema",
 )
